"""Schemas for all of the messages types that we support into the status server."""

from iotile.core.utilities.schema_verify import DictionaryVerifier, IntVerifier, StringVerifier, LiteralVerifier, OptionsVerifier, BooleanVerifier, FloatVerifier

# Commands that we support
HeartbeatCommand = DictionaryVerifier()
HeartbeatCommand.add_required('type', LiteralVerifier('command'))
HeartbeatCommand.add_required('operation', LiteralVerifier('heartbeat'))
HeartbeatCommand.add_required('name', StringVerifier())
HeartbeatCommand.add_required('no_response', BooleanVerifier())

ServiceListCommand = DictionaryVerifier()
ServiceListCommand.add_required('type', LiteralVerifier('command'))
ServiceListCommand.add_required('operation', LiteralVerifier('list_services'))
ServiceListCommand.add_required('no_response', BooleanVerifier())

ServiceQueryCommand = DictionaryVerifier()
ServiceQueryCommand.add_required('type', LiteralVerifier('command'))
ServiceQueryCommand.add_required('operation', LiteralVerifier('query_status'))
ServiceQueryCommand.add_required('name', StringVerifier())
ServiceQueryCommand.add_required('no_response', BooleanVerifier())

QueryMessagesCommand = DictionaryVerifier()
QueryMessagesCommand.add_required('type', LiteralVerifier('command'))
QueryMessagesCommand.add_required('operation', LiteralVerifier('query_messages'))
QueryMessagesCommand.add_required('name', StringVerifier())
QueryMessagesCommand.add_required('no_response', BooleanVerifier())

QueryHeadlineCommand = DictionaryVerifier()
QueryHeadlineCommand.add_required('type', LiteralVerifier('command'))
QueryHeadlineCommand.add_required('operation', LiteralVerifier('query_headline'))
QueryHeadlineCommand.add_required('name', StringVerifier())
QueryHeadlineCommand.add_required('no_response', BooleanVerifier())

RegisterServiceCommand = DictionaryVerifier()
RegisterServiceCommand.add_required('type', LiteralVerifier('command'))
RegisterServiceCommand.add_required('operation', LiteralVerifier('register_service'))
RegisterServiceCommand.add_required('name', StringVerifier())
RegisterServiceCommand.add_required('long_name', StringVerifier())
RegisterServiceCommand.add_required('no_response', BooleanVerifier())

ServiceInfoCommand = DictionaryVerifier()
ServiceInfoCommand.add_required('type', LiteralVerifier('command'))
ServiceInfoCommand.add_required('operation', LiteralVerifier('query_info'))
ServiceInfoCommand.add_required('name', StringVerifier())
ServiceInfoCommand.add_required('no_response', BooleanVerifier())

UpdateStateCommand = DictionaryVerifier()
UpdateStateCommand.add_required('type', LiteralVerifier('command'))
UpdateStateCommand.add_required('operation', LiteralVerifier('update_state'))
UpdateStateCommand.add_required('name', StringVerifier())
UpdateStateCommand.add_required('new_status', IntVerifier())
UpdateStateCommand.add_required('no_response', BooleanVerifier())

PostMessageCommand = DictionaryVerifier()
PostMessageCommand.add_required('type', LiteralVerifier('command'))
PostMessageCommand.add_required('operation', LiteralVerifier('post_message'))
PostMessageCommand.add_required('name', StringVerifier())
PostMessageCommand.add_required('level', IntVerifier())
PostMessageCommand.add_required('message', StringVerifier())
PostMessageCommand.add_required('no_response', BooleanVerifier())

SetHeadlineCommand = DictionaryVerifier()
SetHeadlineCommand.add_required('type', LiteralVerifier('command'))
SetHeadlineCommand.add_required('operation', LiteralVerifier('set_headline'))
SetHeadlineCommand.add_required('name', StringVerifier())
SetHeadlineCommand.add_required('level', IntVerifier())
SetHeadlineCommand.add_required('message', StringVerifier())
SetHeadlineCommand.add_required('created_time', FloatVerifier())
SetHeadlineCommand.add_required('now_time', FloatVerifier())
SetHeadlineCommand.add_required('no_response', BooleanVerifier())

CommandMessage = OptionsVerifier(HeartbeatCommand, SetHeadlineCommand, QueryHeadlineCommand, QueryMessagesCommand, PostMessageCommand, UpdateStateCommand, ServiceInfoCommand, RegisterServiceCommand, ServiceListCommand, ServiceQueryCommand)

# Possible response and notification payloads
ServiceInfoPayload = DictionaryVerifier()
ServiceInfoPayload.add_required('short_name', StringVerifier())
ServiceInfoPayload.add_required('long_name', StringVerifier())
ServiceInfoPayload.add_required('preregistered', BooleanVerifier())

ServiceStatusPayload = DictionaryVerifier()
ServiceStatusPayload.add_required('old_status', IntVerifier())
ServiceStatusPayload.add_required('new_status', IntVerifier())
ServiceStatusPayload.add_required('new_status_string', StringVerifier())

MessagePayload = DictionaryVerifier()
MessagePayload.add_required('level', IntVerifier())
MessagePayload.add_required('message', StringVerifier())
MessagePayload.add_required('created_time', FloatVerifier())
MessagePayload.add_required('now_time', FloatVerifier())
MessagePayload.add_optional('count', IntVerifier())
MessagePayload.add_optional('id', IntVerifier())

# Notifications that the SupervisorService can push
ServiceStatusChanged = DictionaryVerifier()
ServiceStatusChanged.add_required('type', LiteralVerifier('notification'))
ServiceStatusChanged.add_required('operation', LiteralVerifier('state_change'))
ServiceStatusChanged.add_required('name', StringVerifier())
ServiceStatusChanged.add_required('payload', ServiceStatusPayload)

ServiceAdded = DictionaryVerifier()
ServiceAdded.add_required('type', LiteralVerifier('notification'))
ServiceAdded.add_required('operation', LiteralVerifier('new_service'))
ServiceAdded.add_required('name', StringVerifier())
ServiceAdded.add_required('payload', ServiceInfoPayload)

HeartbeatReceived = DictionaryVerifier()
HeartbeatReceived.add_required('type', LiteralVerifier('notification'))
HeartbeatReceived.add_required('operation', LiteralVerifier('heartbeat'))
HeartbeatReceived.add_required('name', StringVerifier())

NewMessage = DictionaryVerifier()
NewMessage.add_required('type', LiteralVerifier('notification'))
NewMessage.add_required('operation', LiteralVerifier('new_message'))
NewMessage.add_required('name', StringVerifier())
NewMessage.add_required('payload', MessagePayload)

NewHeadline = DictionaryVerifier()
NewHeadline.add_required('type', LiteralVerifier('notification'))
NewHeadline.add_required('operation', LiteralVerifier('new_headline'))
NewHeadline.add_required('name', StringVerifier())
NewHeadline.add_required('payload', MessagePayload)



"""Unit tests for service message caching."""

import pytest
import iotilegateway.supervisor.states as states
from iotile.core.exceptions import ArgumentError


@pytest.fixture
def service():
    serv = states.ServiceState('test', 'A test service', False, int_id=5, max_messages=10)
    return serv


def test_creating_service():
    """Make sure we can create a service."""

    serv = states.ServiceState('test', 'A test service', False, int_id=5, max_messages=10)

    assert serv.short_name == 'test'
    assert serv.long_name == 'A test service'
    assert serv.preregistered is False
    assert serv.id == 5
    assert serv.state == states.UNKNOWN


def test_posting_messages(service):
    """Make sure we can post messages and read them back."""

    service.post_message(states.INFO_LEVEL, 'test message')
    service.post_message(states.WARNING_LEVEL, 'test 2 message')

    assert len(service.messages) == 2
    assert service.messages[0].level == states.INFO_LEVEL
    assert service.messages[0].message == 'test message'
    assert service.messages[0].count == 1
    assert service.messages[0].id == 0

    assert service.messages[1].level == states.WARNING_LEVEL
    assert service.messages[1].message == 'test 2 message'
    assert service.messages[1].count == 1
    assert service.messages[1].id == 1


def test_duplicate_messages(service):
    """Make sure duplicate messages are deduplicated."""

    service.post_message(states.INFO_LEVEL, 'test message')
    service.post_message(states.INFO_LEVEL, 'test message')

    assert len(service.messages) == 1
    assert service.messages[0].level == states.INFO_LEVEL
    assert service.messages[0].message == 'test message'
    assert service.messages[0].count == 2
    assert service.messages[0].id == 0


def test_message_creation():
    """Make sure messages can be serialized and deserialized."""

    ts = 1000.0
    now = 1010.0

    msg = states.ServiceMessage(states.ERROR_LEVEL, 'test message', 15, ts, now)

    msg_dict = msg.to_dict()
    new_msg = states.ServiceMessage.FromDictionary(msg_dict)

    msg_age = msg_dict['now_time'] - msg_dict['created_time']
    del msg_dict['created_time']
    del msg_dict['now_time']

    new_dict = new_msg.to_dict()
    new_age = new_dict['now_time'] - new_dict['created_time']
    del new_dict['created_time']
    del new_dict['now_time']

    # The message ages should be identical to within a few lsbs of a float but
    # technically we don't really care how precise it is
    assert abs(new_age - msg_age) < 0.1
    assert msg_dict == new_dict
    assert new_msg.level == states.ERROR_LEVEL
    assert new_msg.message == 'test message'
    assert new_msg.id == 15
    assert new_msg.count == 1



"""iotile-gateway script main entry point."""

import signal
import logging
import json
from iotilegateway.gateway import IOTileGateway
from tornado.options import define, parse_command_line, options
from iotile.core.exceptions import ArgumentError

define('config', help="Config file for defining what adapters and agents to use")


def main():
    """Main entry point for iotile-gateway."""

    gateway = None
    logging.basicConfig(format='%(levelname)-.1s-%(asctime)-15s-%(module)-10s:%(lineno)-4s %(message)s')
    log = logging.getLogger(__name__)

    def quit_signal_handler(signum, frame):  # pylint: disable=W0613
        """Signal handler to catch ^C and cleanly shut down."""

        log.critical("In quit signal handler.")

        if gateway is not None:
            log.critical("Calling stop on gateway loop")
            gateway.stop_from_signal()

    try:
        parse_command_line()

        config_file = options.config
        if config_file is None:
            log.critical("You must pass a config file using --config=<path to file>")
            return 1

        try:
            with open(config_file, "rb") as conf:
                args = json.load(conf)
        except IOError, exc:
            raise ArgumentError("Could not open required config file", path=config_file, error=str(exc))
        except ValueError, exc:
            raise ArgumentError("Could not parse JSON from config file", path=config_file, error=str(exc))
        except TypeError, exc:
            raise ArgumentError("You must pass the path to a json config file", path=config_file)

        signal.signal(signal.SIGINT, quit_signal_handler)

        gateway = IOTileGateway(args)
        gateway.start()
        gateway.wait()

    except Exception:  # pylint: disable=W0703
        log.exception("Fatal error starting gateway")



"""Class for managing logs and state from various services."""

from monotonic import monotonic
from iotile.core.exceptions import ArgumentError
import states


class ServiceManager:
    """A simple repository for handling the state of a running service and querying log messages from it."""

    def __init__(self, expected_services):
        """Constructor.

        Args:
            expected_services (dict): A dictionary with the name of expected services that should be running
                other services may register on the fly but expected services will allow reporting of a
                NOT_STARTED status and querying the number of services that should be running.
        """
        self.services = {}
        self._monitors = set()

        for service in expected_services:
            self.add_service(service['short_name'], service['long_name'], preregistered=True)

    def add_monitor(self, callback):
        """Register a callback whenever a service changes.

        Args:
            callback (callable): A callback with signature callback(short_name, update)
                where update is a dictionary with the keys:
                    type: one of state_change, heartbeat, new_log, new_message, new_service
                    info: optional dictionary that depends on the type of the message
        """

        self._monitors.add(callback)

    def remove_monitor(self, callback):
        """Remove a previously registered callback.

        Args:
            callback (callable): A callback that was previously registered
        """

        try:
            self._monitors.remove(callback)
        except KeyError:
            raise ArgumentError("Callback was not registered")

    def _notify_update(self, name, change_type, change_info=None):
        """Notify updates on a service to anyone who cares."""

        for monitor in self._monitors:
            try:
                monitor(name, change_type, change_info)
            except Exception:
                # We can't allow any exceptions in a monitor routine to break the server.
                pass

    def update_state(self, short_name, state):
        """Set the current state of a service.

        If the state is unchanged from a previous attempt, this routine does
        nothing.

        Args:
            short_name (string): The short name of the service
            state (int): The new stae of the service
        """

        if short_name not in self.services:
            raise ArgumentError("Service name is unknown", short_name=short_name)

        if state not in states.KNOWN_STATES:
            raise ArgumentError("Invalid service state", state=state)

        serv = self.services[short_name]['state']

        if serv.state == state:
            return

        update = {}
        update['old_status'] = serv.state
        update['new_status'] = state
        update['new_status_string'] = states.KNOWN_STATES[state]

        serv.state = state
        self._notify_update(short_name, 'state_change', update)

    def add_service(self, short_name, long_name, preregistered=False):
        """Add a service to the list of tracked services.

        Args:
            short_name (string): A unique short service name for the service
            long_name (string): A longer, user friendly name for the service
            preregistered (bool): Whether this service is an expected preregistered
                service.
        """

        if short_name in self.services:
            raise ArgumentError("Could not add service because the short_name is taken", short_name=short_name)

        serv_state = states.ServiceState(short_name, long_name, preregistered)

        service = {
            'state': serv_state,
            'heartbeat_threshold': 600
        }

        self.services[short_name] = service
        self._notify_update(short_name, 'new_service', self.service_info(short_name))

    def service_info(self, short_name):
        """Get static information about a service.

        Args:
            short_name (string): The short name of the service to query

        Returns:
            dict: A dictionary with the long_name and preregistered info
                on this service.
        """

        if short_name not in self.services:
            raise ArgumentError("Unknown service name", short_name=short_name)

        info = {}
        info['short_name'] = short_name
        info['long_name'] = self.services[short_name]['state'].long_name
        info['preregistered'] = self.services[short_name]['state'].preregistered

        return info

    def service_messages(self, short_name):
        """Get the messages stored for a service.

        Args:
            short_name (string): The short name of the service to get messages for

        Returns:
            list(ServiceMessage): A list of the ServiceMessages stored for this service
        """

        if short_name not in self.services:
            raise ArgumentError("Unknown service name", short_name=short_name)

        return self.services[short_name]['state'].messages

    def service_headline(self, short_name):
        """Get the headline stored for a service.

        Args:
            short_name (string): The short name of the service to get messages for

        Returns:
            ServiceMessage: the headline or None if there is no headline
        """

        if short_name not in self.services:
            raise ArgumentError("Unknown service name", short_name=short_name)

        return self.services[short_name]['state'].headline

    def service_status(self, short_name):
        """Get the current status of a service.

        Returns information about the service such as the length since the last
        heartbeat, any status messages that have been posted about the service
        and whether the heartbeat should be considered out of the ordinary.

        Args:
            short_name (string): The short name of the service to query

        Returns:
            dict: A dictionary with the status of the service
        """

        if short_name not in self.services:
            raise ArgumentError("Unknown service name", short_name=short_name)

        info = {}

        service = self.services[short_name]['state']

        info['heartbeat_age'] = monotonic() - service.last_heartbeat
        info['numeric_status'] = service.state
        info['string_status'] = service.string_state

        return info

    def send_message(self, short_name, level, message):
        """Post a message for a service.

        Args:
            short_name (string): The short name of the service to query
            level (int): The level of the message (info, warning, error)
            message (string): The message contents
        """

        if short_name not in self.services:
            raise ArgumentError("Unknown service name", short_name=short_name)

        now = monotonic()

        self.services[short_name]['state'].post_message(level, message)
        self._notify_update(short_name, 'new_message', {'level': level, 'message': message, 'created_time': now, 'now_time': now})

    def set_headline(self, short_name, level, message):
        """Set the sticky headline for a service.

        Args:
            short_name (string): The short name of the service to query
            level (int): The level of the message (info, warning, error)
            message (string): The message contents
        """

        if short_name not in self.services:
            raise ArgumentError("Unknown service name", short_name=short_name)

        now = monotonic()

        self.services[short_name]['state'].set_headline(level, message)
        self._notify_update(short_name, 'new_headline', {'level': level, 'message': message, 'created_time': now, 'now_time': now})

    def send_heartbeat(self, short_name):
        """Post a heartbeat for a service.

        Args:
            short_name (string): The short name of the service to query
        """

        if short_name not in self.services:
            raise ArgumentError("Unknown service name", short_name=short_name)

        self.services[short_name]['state'].heartbeat()
        self._notify_update(short_name, 'heartbeat')

    def list_services(self):
        """Get a list of the services known to this ServiceManager.

        Returns:
            list(string): A list of string short names for the known services
        """

        return self.services.keys()



"""All possible states that a service can be in."""

from iotile.core.exceptions import ArgumentError
from monotonic import monotonic
from collections import deque
from command_formats import MessagePayload

# Service states
NOT_STARTED = 0
RUNNING = 1
DEGRADED = 2
STOPPED = 3
UNKNOWN = 4

KNOWN_STATES = {
    NOT_STARTED: 'Not Started',
    RUNNING: 'Running',
    STOPPED: 'Stopped',
    DEGRADED: 'Degraded',
    UNKNOWN: 'Unknown'
}

# Message Levels
INFO_LEVEL = 0
WARNING_LEVEL = 1
ERROR_LEVEL = 2


class ServiceMessage(object):
    """A message received from a service."""

    def __init__(self, level, message, message_id, timestamp=None, now_reference=None):
        """Constructor.

        Args:
            level (int): The message importance
            message (string): The message contents
            message_id (int): A unique id for the message
            timestamp (float): An optional monotonic value in seconds for when the message was created
            now_reference (float): If timestamp is not relative to monotonic() as called from this
                module then this should be now() as seen by whoever created the timestamp.
        """

        self.level = level
        self.message = message
        self.count = 1
        self.id = message_id

        if timestamp is None:
            self.created = monotonic()
        elif now_reference is None:
            self.created = timestamp
        else:
            # Figure out the age of the timestamp relative to the now reference from the creator
            # and subtract that age from now for us to convert the timestamp into our frame of
            # reference.
            now = monotonic()
            adj = now - now_reference
            self.created = timestamp + adj

            # Make sure we never create times in the future so that age below is always nonnegative
            if self.created > now:
                self.created = now

    @classmethod
    def FromDictionary(cls, msg_dict):
        """Create from a dictionary with kv pairs.

        Args:
            msg_dict (dict): A dictionary with information as created by to_dict()

        Returns:
            ServiceMessage: the converted message
        """

        MessagePayload.verify(msg_dict)

        level = msg_dict.get('level')
        msg = msg_dict.get('message')
        now = msg_dict.get('now_time')
        created = msg_dict.get('created_time')
        count = msg_dict.get('count', 1)
        msg_id = msg_dict.get('id', 0)

        new_msg = ServiceMessage(level, msg, msg_id, created, now)
        if count > 1:
            new_msg.count = count

        return new_msg

    def to_dict(self):
        """Create a dictionary with the information in this message.

        Returns:
            dict: The dictionary with information
        """

        msg_dict = {}
        msg_dict['level'] = self.level
        msg_dict['message'] = self.message
        msg_dict['now_time'] = monotonic()
        msg_dict['created_time'] = self.created
        msg_dict['id'] = self.id
        msg_dict['count'] = self.count

        return msg_dict

    @property
    def age(self):
        """Return the age of this message in seconds.

        Returns:
            float: the age in seconds
        """

        return monotonic() - self.created


class ServiceState(object):
    """A simple object for holding the current state of a service."""

    def __init__(self, short_name, long_name, preregistered, int_id=None, max_messages=5):
        """Constructor.

        Args:
            short_name (string): A unique short name for the service
            long_name (string): A user friendly name for the service
            preregistered (bool): Whether this is an expected preregistered
                service
            int_id (int): An internal numeric id for this service
            max_messages (int): The maximum number of messages to keep
        """

        self.short_name = short_name
        self.long_name = long_name
        self.preregistered = preregistered
        self.last_heartbeat = monotonic()
        self.num_heartbeats = 0
        self.id = int_id
        self._state = UNKNOWN
        self.messages = deque(maxlen=max_messages)
        self.headline = None
        self._last_message_id = 0

    @property
    def string_state(self):
        """A string for the curent state."""
        return KNOWN_STATES[self._state]

    @property
    def state(self):
        """The current numeric service state."""
        return self._state

    @property
    def heartbeat_age(self):
        """The time in seconds since the last heartbeat."""
        return monotonic() - self.last_heartbeat

    @state.setter
    def state(self, new_state):
        if new_state not in KNOWN_STATES:
            raise ArgumentError("Unknown service state", state=new_state)

        self._state = new_state

    def get_message(self, message_id):
        """Get a message by its persistent id.

        Args:
            message_id (int): The id of the message that we're looking for
        """

        for message in self.messages:
            if message.id == message_id:
                return message

        raise ArgumentError("Message ID not found", message_id=message_id)

    def post_message(self, level, message, count=1, timestamp=None, now_reference=None):
        """Post a new message for service.

        Args:
            level (int): The level of the message (info, warning, error)
            message (string): The message contents
            count (int): The number of times the message has been repeated
            timestamp (float): An optional monotonic value in seconds for when the message was created
            now_reference (float): If timestamp is not relative to monotonic() as called from this
                module then this should be now() as seen by whoever created the timestamp.
        """

        if len(self.messages) > 0 and self.messages[-1].message == message:
            self.messages[-1].count += 1
        else:
            msg_object = ServiceMessage(level, message, self._last_message_id, timestamp, now_reference)
            msg_object.count = count
            self.messages.append(msg_object)
            self._last_message_id += 1

    def set_headline(self, level, message, timestamp=None, now_reference=None):
        """Set the persistent headline message for this service.

        Args:
            level (int): The level of the message (info, warning, error)
            message (string): The message contents
            timestamp (float): An optional monotonic value in seconds for when the message was created
            now_reference (float): If timestamp is not relative to monotonic() as called from this
                module then this should be now() as seen by whoever created the timestamp.
        """

        if self.headline is not None and self.headline.message == message:
            self.headline.created = monotonic()
            self.headline.count += 1
            return

        msg_object = ServiceMessage(level, message, self._last_message_id, timestamp, now_reference)
        self.headline = msg_object
        self._last_message_id += 1

    def heartbeat(self):
        """Record a heartbeat for this service."""

        self.last_heartbeat = monotonic()
        self.num_heartbeats += 1



"""Unit tests for supervisor server and client."""

import pytest
import logging
from iotilegateway.supervisor.ws_handler import ServiceWebSocketHandler
from iotilegateway.supervisor.service_manager import ServiceManager
from iotilegateway.supervisor.status_client import ServiceStatusClient
import iotilegateway.supervisor.states as states
import tornado.gen
import tornado.testing
from util_async import AsyncWebSocketsTestCase
from iotile.core.exceptions import ArgumentError


class TestSupervisor(AsyncWebSocketsTestCase):
    """Test the supervisor service and websocket synchronizing client."""

    def _initialize(self):
        services = [{'short_name': 'service1', 'long_name': 'Service 1'}, {'short_name': 'service2', 'long_name': 'Service 2'}]
        self.manager = ServiceManager(services)
        self.client = None

    @tornado.concurrent.run_on_executor
    def _preshutdown_deinitialize(self):
        if self.client is not None:
            self.client.stop()

    def _get_app(self):
        app = tornado.web.Application([
            (r'/services', ServiceWebSocketHandler, {'manager': self.manager, 'logger': logging.getLogger('test.logger')})
        ])

        return app

    @tornado.concurrent.run_on_executor
    def _get_client(self):
        self.client = ServiceStatusClient(self.get_supervisor_port())
        return self.client

    @tornado.concurrent.run_on_executor
    def _list_services(self):
        servs = self.client.list_services()
        return servs

    @tornado.concurrent.run_on_executor
    def _query_service(self, name):
        status = self.client.service_status(name)
        return status

    @tornado.concurrent.run_on_executor
    def _query_info(self, name):
        status = self.client.service_info(name)
        return status

    @tornado.concurrent.run_on_executor
    def _heartbeat(self, name):
        self.client.send_heartbeat(name)

    @tornado.concurrent.run_on_executor
    def _register_service(self, short_name, long_name):
        status = self.client.register_service(short_name, long_name)
        return status

    @tornado.concurrent.run_on_executor
    def _get_messages(self, short_name):
        msgs = self.client.get_messages(short_name)
        return msgs

    @tornado.concurrent.run_on_executor
    def _post_error(self, short_name, msg):
        self.client.post_error(short_name, msg)

    @tornado.concurrent.run_on_executor
    def _update_state(self, short_name, state):
        self.client.update_state(short_name, state)

    @tornado.concurrent.run_on_executor
    def _set_headline(self, short_name, level, message):
        self.client.post_headline(short_name, level, message)

    @tornado.concurrent.run_on_executor
    def _get_headline(self, short_name):
        return self.client.get_headline(short_name)

    @tornado.testing.gen_test
    def test_list_services(self):
        """Make sure we can pull a service list from the supervisor."""

        yield self._get_client()
        servs = yield self._list_services()

        assert len(servs) == 2
        assert 'service1' in servs
        assert 'service2' in servs

    @tornado.testing.gen_test
    def test_query_status(self):
        """Make sure we can query a service's status."""

        yield self._get_client()
        status = yield self._query_service('service2')

        assert status['numeric_status'] == states.UNKNOWN

        with pytest.raises(ArgumentError):
            yield self._query_service('service3')

    @tornado.testing.gen_test
    def test_register_service(self):
        """Make sure we can register a new service."""

        yield self._get_client()
        yield self._register_service('service3', 'A nice service')

        servs = yield self._list_services()
        assert len(servs) == 3
        assert 'service3' in servs

    @tornado.testing.gen_test
    def test_query_info(self):
        """Make sure we can register a new service."""

        yield self._get_client()
        status = yield self._query_info('service2')
        assert status['long_name'] == 'Service 2'
        assert status['preregistered'] is True

        yield self._register_service('service3', 'A nice service')
        status = yield self._query_info('service3')
        assert status['long_name'] == 'A nice service'
        assert status['preregistered'] is False

    @tornado.testing.gen_test
    def test_service_syncing(self):
        """Make sure we get updates on service changes."""

        yield self._get_client()
        yield self._register_service('service3', 'A nice service')

        # Make sure the update got synced
        yield self._query_info('service3')

        assert len(self.client.services) == 3
        assert 'service3' in self.client.services

        serv = self.client.services['service3']
        assert serv.long_name == 'A nice service'
        assert serv.preregistered is False
        assert serv.id == 2

        # Update the state of a service and make sure it gets synced
        yield self._update_state('service3', states.RUNNING)
        yield self._query_info('service3')

        assert self.client.services['service3'].state == states.RUNNING
        assert self.client.services['service3'].string_state == states.KNOWN_STATES[states.RUNNING]

    @tornado.testing.gen_test
    def test_service_heartbeat(self):
        """Make sure we get updates on service heartbeats."""

        yield self._get_client()

        assert self.client.services['service1'].num_heartbeats == 0

        yield self._heartbeat('service1')
        yield self._query_info('service1')
        assert self.client.services['service1'].num_heartbeats == 1

        yield self._heartbeat('service1')
        yield self._query_info('service1')
        assert self.client.services['service1'].num_heartbeats == 2

    @tornado.testing.gen_test
    def test_query_messages(self):
        """Make sure we can set, sync and query messages from the supervisor."""

        yield self._get_client()

        msgs = yield self._get_messages('service1')
        assert len(msgs) == 0

        yield self._post_error('service1', 'test 1')
        yield self._post_error('service1', 'test 1')
        yield self._post_error('service1', 'test 2')
        yield self._post_error('service1', 'test 3')
        yield self._post_error('service1', 'test 2')

        msgs = yield self._get_messages('service1')
        assert len(msgs) == 4
        assert msgs[0].count == 2
        assert msgs[0].message == 'test 1'
        assert msgs[1].count == 1
        assert msgs[1].message == 'test 2'
        assert msgs[2].count == 1
        assert msgs[2].message == 'test 3'
        assert msgs[3].count == 1
        assert msgs[3].message == 'test 2'

        # Now make sure the messages got properly synced locally as well
        local = self.client.local_service('service1')
        msgs = local.messages
        assert len(msgs) == 4
        assert msgs[0].count == 2
        assert msgs[0].message == 'test 1'
        assert msgs[1].count == 1
        assert msgs[1].message == 'test 2'
        assert msgs[2].count == 1
        assert msgs[2].message == 'test 3'
        assert msgs[3].count == 1
        assert msgs[3].message == 'test 2'

    @tornado.testing.gen_test
    def test_service_headline(self):
        """Make sure we can set, sync and query headlines."""

        yield self._get_client()

        msg = yield self._get_headline('service1')
        assert msg is None

        yield self._set_headline('service1', states.ERROR_LEVEL, 'test message')
        msg = yield self._get_headline('service1')

        assert msg.level == states.ERROR_LEVEL
        assert msg.message == 'test message'
        assert msg.count == 1

        local = self.client.local_service('service1')
        assert local.headline is not None
        assert local.headline.message == 'test message'

        # Make sure we can change the headline
        yield self._set_headline('service1', states.INFO_LEVEL, 'info message')
        msg = yield self._get_headline('service1')
        assert msg.level == states.INFO_LEVEL
        assert msg.message == 'info message'

        local = self.client.local_service('service1')
        assert local.headline is not None
        assert local.headline.message == 'info message'



import unittest
import pytest
import os.path
import os
import struct
import logging
import sys
from iotile.mock.mock_iotile import MockIOTileDevice
from iotile.mock.mock_adapter import MockDeviceAdapter
from iotile.core.hw.reports.individual_format import IndividualReadingReport
from iotile.core.hw.reports.report import IOTileReading
from iotilegateway.device import DeviceManager
import threading
from tornado.ioloop import IOLoop
import tornado.gen
import tornado.testing

class TestDeviceManager(tornado.testing.AsyncTestCase):
    def setUp(self):
        super(TestDeviceManager, self).setUp()

        self.dev = MockIOTileDevice(1, 'TestCN')
        self.dev.reports = [IndividualReadingReport.FromReadings(100, [IOTileReading(0, 1, 2)])]
        self.adapter = MockDeviceAdapter()
        self.adapter.add_device('test', self.dev)
        
        self.manager = DeviceManager(self.io_loop)
        self.manager.add_adapter(self.adapter)
        self.manager.register_monitor(1, ['report'], self.on_report)
        self.reports = []
        self.reports_received = threading.Event()

    def tearDown(self):
        super(TestDeviceManager, self).tearDown()

    def on_report(self, dev_uuid, event_name, report):
        """Callback triggered when a report is received from a device on a device adapter
        """

        self.reports.append(report)
        self.reports_received.set()

    @tornado.testing.gen_test
    def test_connect_direct(self):
        """Make sure we can directly connect to a device
        """

        res = yield self.manager.connect_direct('0/test')
        assert res['success'] is True
    
    @tornado.testing.gen_test
    def test_send_rpc(self):
        res = yield self.manager.connect_direct('0/test')
        assert res['success'] is True

        conn_id = res['connection_id']

        res = yield self.manager.send_rpc(conn_id, 8, 0, 4, '', 1.0) 
        assert len(res['payload']) == 6
        assert res['payload'] == 'TestCN'

    def test_monitors(self):
        mon_id = self.manager.register_monitor(10, ['report'], lambda x,y: x)
        self.manager.adjust_monitor(mon_id, add_events=['connection'], remove_events=['report'])
        self.manager.remove_monitor(mon_id)

    def test_scan(self):
        devs = self.manager.scanned_devices
        assert len(devs) == 0

        self.adapter.advertise()

        #Let the ioloop process the advertisements
        try:
            self.wait(timeout=0.1)
        except:
            pass

        devs = self.manager.scanned_devices
        assert len(devs) == 1
        assert 1 in devs

    @tornado.testing.gen_test
    def test_connect(self):
        """Make sure we can directly to a device by uuid
        """

        self.adapter.advertise()
        
        yield tornado.gen.sleep(0.1)

        print self.manager.scanned_devices

        res = yield self.manager.connect(1)
        print res

        assert res['success'] is True

    @tornado.testing.gen_test
    def test_reports(self):
        self.adapter.advertise()
        yield tornado.gen.sleep(0.1)

        res = yield self.manager.connect(1)
        conn_id = res['connection_id']

        yield self.manager.open_interface(conn_id, 'streaming')
        yield tornado.gen.sleep(0.1)

        assert len(self.reports) == 1
        print self.reports[0]



"""An IOTile gateway-in-a-box that will connect to devices using device adapters and serve them using agents."""

import logging
import threading
import pkg_resources
import tornado.ioloop
from iotilegateway.supervisor import ServiceStatusClient
import iotilegateway.supervisor.states as states
import iotilegateway.device as device
from iotile.core.exceptions import ArgumentError


def find_entry_point(group, name):
    """Find an entry point by name.

    Args:
        group (string): The entry point group like iotile.gateway_agent
        name (string): The name of the entry point to find
    """

    for entry in pkg_resources.iter_entry_points(group, name):
        item = entry.load()
        return item

    raise ArgumentError("Could not find installed plugin by name and group", group=group, name=name)


class IOTileGateway(threading.Thread):
    """A gateway that finds IOTile devices using device adapters and serves them using agents.

    The gateway runs in separate thread in a tornado IOLoop and you can call the synchronous
    wait function to wait for it to quit.  It will loop forever unless you stop it by calling
    the stop() or stop_from_signal() methods.  These functions add a task to the gateway's
    event loop and implicitly call wait to synchronously wait until the gateway loop actually
    stops.

    Args:
        config (dict): The configuration of the gateway.  There should be two keys set:
            agents: a list of dictionaries with the name of the agent and any arguments that
                should be passed to create it.
            adapters: a list of dictionaries with the device adapters to add into the gateway
                and any argument sthat should be use to create each one.
    """

    def __init__(self, config):
        self.loop = tornado.ioloop.IOLoop.instance()
        self.device_manager = device.DeviceManager(self.loop)
        self.agents = []
        self.supervisor = None
        self.loaded = threading.Event()

        self._config = config
        self._logger = logging.getLogger(__name__)

        if 'agents' not in config:
            self._config['agents'] = []
            self._logger.warn("No agents defined in arguments to iotile-gateway, this is likely not what you want")
        elif 'adapters' not in config:
            self._config['adapters'] = []
            self._logger.warn("No device adapters defined in arguments to iotile-gateway, this is likely not what you want")

        super(IOTileGateway, self).__init__()

    def run(self):
        """Start the gateway and run it to completion in another thread."""

        # If we have an initialization error, stop trying to initialize more things and
        # just shut down cleanly
        should_close = False

        # Load in all of the gateway agents that are supposed to provide access to
        # the devices in this gateway
        for agent_info in self._config['agents']:
            if 'name' not in agent_info:
                self._logger.error("Invalid agent information in gateway config, info=%s, missing_key=%s", str(agent_info), 'name')
                should_close = True
                break

            agent_name = agent_info['name']
            agent_args = agent_info.get('args', {})

            self._logger.info("Loading agent by name '%s'", agent_name)
            agent_class = find_entry_point('iotile.gateway_agent', agent_name)
            try:
                agent = agent_class(agent_args, self.device_manager, self.loop)
                agent.start()
                self.agents.append(agent)
            except Exception:  # pylint: disable=W0703
                self._logger.exception("Could not load gateway agent %s, quitting", agent_name)
                should_close = True
                break

        # Load in all of the device adapters that provide access to actual devices
        if not should_close:
            for adapter_info in self._config['adapters']:
                if 'name' not in adapter_info:
                    self._logger.error("Invalid adapter information in gateway config, info=%s, missing_key=%s", str(adapter_info), 'name')
                    should_close = True
                    break

                adapter_name = adapter_info['name']
                port_string = adapter_info.get('port', None)

                self._logger.info("Loading device adapter by name '%s' and port '%s'", adapter_name, port_string)

                try:
                    adapter_class = find_entry_point('iotile.device_adapter', adapter_name)
                    adapter = adapter_class(port_string)
                    self.device_manager.add_adapter(adapter)
                except Exception:  # pylint: disable=W0703
                    self._logger.exception("Could not load device adapter %s, quitting", adapter_name)
                    should_close = True

        if should_close:
            self.loop.add_callback(self._stop_loop)
        else:
            # Notify that we have now loaded all plugins and are starting operation
            self.loaded.set()
            # Try to regularly update a supervisor about our status
            callback = tornado.ioloop.PeriodicCallback(self._try_report_status, 60000)
            callback.start()

        self.loop.start()

        # The loop has been closed, finish and quit
        self._logger.critical("Done stopping loop")

    def _try_report_status(self):
        """Periodic callback to report our gateway's status."""

        if self.supervisor is None:
            try:
                self.supervisor = ServiceStatusClient('ws://localhost:9400/services')
                self.supervisor.register_service('gateway', 'Device Gateway')
                self.supervisor.post_info('gateway', "Service started successfully")
                self.supervisor.post_headline('gateway', states.INFO_LEVEL, 'Started successfully')
            except Exception:  # pylint: disable=W0703
                self._logger.exception("Exception trying to create a ServiceStatusClient")
                return

        self.supervisor.update_state('gateway', states.RUNNING)
        self.supervisor.send_heartbeat('gateway')

    def _stop_loop(self):
        """Cleanly stop the gateway and close down the IOLoop.

        This function must be called only by being added to our event loop using add_callback.
        """

        self._logger.critical("Stopping gateway")
        self._logger.info("Stopping gateway agents")

        for agent in self.agents:
            try:
                agent.stop()
            except Exception:  # pylint: disable=W0703
                self._logger.exception("Error stopping gateway agent")

        self._logger.critical('Stopping device adapters')

        try:
            self.device_manager.stop()
        except Exception:  # pylint: disable=W0703
            self._logger.exception("Error stopping device adapters")

        if self.supervisor:
            try:
                self.supervisor.update_state('gateway', states.STOPPED)
                self.supervisor.post_headline('gateway', states.INFO_LEVEL, 'Stoppped by supervisor')
            except Exception:  # pylint: disable=W0703
                self._logger.exception("Error updating service status to stopped")

            try:
                self.supervisor.stop()
            except Exception:  # pylint: disable=W0703
                self._logger.exception("Error stopping IOLoop")

        self.loop.stop()
        self._logger.critical('Stopping event loop and shutting down')

    def stop(self):
        """Stop the gateway manager and synchronously wait for it to stop."""

        self.loop.add_callback(self._stop_loop)
        self.wait()

    def wait(self):
        """Wait for this gateway to shut down.

        We need this special function because waiting inside
        join will cause signals to not get handled.
        """

        while self.is_alive():
            try:
                self.join(timeout=0.1)
            except IOError:
                pass  # IOError comes when this call is interrupted in a signal handler

    def stop_from_signal(self):
        """Stop the gateway from a signal handler, not waiting for it to stop."""

        self.loop.add_callback_from_signal(self._stop_loop)



import time
from iotile.mock.mock_iotile import MockIOTileDevice
from iotile.mock.mock_adapter import MockDeviceAdapter
from iotile.core.hw.hwmanager import HardwareManager
from iotilegateway.device import DeviceManager
from iotilegateway.wshandler import WebSocketHandler
from iotile.core.hw.reports.individual_format import IndividualReadingReport
from iotile.core.hw.reports.report import IOTileReading
import tornado.gen
import tornado.testing
from util_async import AsyncWebSocketsTestCase


class TestWebSocketInterface(AsyncWebSocketsTestCase):
    def _initialize(self):
        self.dev = MockIOTileDevice(1, 'TestCN')
        self.dev.reports = [IndividualReadingReport.FromReadings(100, [IOTileReading(0, 1, 2)])]
        self.adapter = MockDeviceAdapter()
        self.adapter.add_device('test', self.dev)

        self.manager = DeviceManager(self.io_loop)
        self.manager.add_adapter(self.adapter)
        self.hw = None

    def _deinitialize(self):
        pass

    def _get_app(self):
        app = tornado.web.Application([
            (r'/iotile/v1', WebSocketHandler, {'manager': self.manager})
        ])

        return app

    @tornado.gen.coroutine
    def ensure_advertised(self):
        self.adapter.advertise()

        yield tornado.gen.sleep(0.1)
        devs = self.manager.scanned_devices
        assert len(devs) == 1
        assert 1 in devs

    @tornado.concurrent.run_on_executor
    def get_hwmanager(self):
        hw = HardwareManager(port=self.get_iotile_port())
        return hw

    @tornado.concurrent.run_on_executor
    def connect(self, uuid):
        self.hw.connect(uuid)

    @tornado.concurrent.run_on_executor
    def enable_streaming(self):
        self.hw.enable_streaming()

    @tornado.testing.gen_test
    def test_basic_connection(self):
        yield self.ensure_advertised()

        self.hw = yield self.get_hwmanager()
        yield self.connect(1)

    @tornado.testing.gen_test
    def test_reports(self):
        yield self.ensure_advertised()

        self.hw = yield self.get_hwmanager()
        yield self.connect(1)

        assert self.hw.count_reports() == 0
        yield self.enable_streaming()

        #Give time for report to be processed
        time.sleep(0.1)

        assert self.hw.count_reports() == 1



import signal
import logging
import sys
import tornado.ioloop
import json
from tornado.options import define, parse_command_line, options

from iotile.core.exceptions import ArgumentError
from ws_handler import ServiceWebSocketHandler
from service_manager import ServiceManager

should_close = False

define('config', help="Optional config file for defining expected services and our port for connections")


def quit_signal_handler(signum, frame):
    """Signal handler to catch ^C and cleanly shut down."""

    global should_close

    should_close = True
    log = logging.getLogger('tornado.general')
    log.critical('Received stop signal, attempting to stop')


def try_quit():
    """Periodic callback to attempt to cleanly shut down this gateway."""
    global should_close

    if not should_close:
        return

    log = logging.getLogger('tornado.general')

    log.critical("Stopping Supervisor Service")

    tornado.ioloop.IOLoop.instance().stop()
    log.critical('Stopping event loop and shutting down')


def main():
    """Main entry point for iotile-supervisor."""
    global should_close

    log = logging.getLogger('tornado.general')

    parse_command_line()

    config_file = options.config
    args = {}

    if config_file is not None:
        try:
            with open(config_file, "rb") as conf:
                args = json.load(conf)
        except IOError, exc:
            raise ArgumentError("Could not open required config file", path=config_file, error=str(exc))
        except ValueError, exc:
            raise ArgumentError("Could not parse JSON from config file", path=config_file, error=str(exc))
        except TypeError, exc:
            raise ArgumentError("You must pass the path to a json config file", path=config_file)

    if 'port' not in args:
        args['port'] = 9400

    if 'expected_services' not in args:
        args['expected_services'] = {}

    service_manager = ServiceManager(args['expected_services'])

    loop = tornado.ioloop.IOLoop.instance()

    signal.signal(signal.SIGINT, quit_signal_handler)

    # Make sure we have a way to cleanly break out of the event loop on Ctrl-C
    tornado.ioloop.PeriodicCallback(try_quit, 100).start()

    port = args.get('port')

    app = tornado.web.Application([
        (r'/services', ServiceWebSocketHandler, {'manager': service_manager, 'logger': log}),
    ])

    log.info("Starting IOTile supervisor service over websockets on port %d" % port)
    app.listen(port)

    loop.start()

    # The loop has been closed, finish and quit
    log.critical("Done stopping loop")



import tornado.web
import logging
from wshandler import WebSocketHandler

class WebSocketGatewayAgent(object):
    """A gateway agent for connecting to an IOTile gateway over websockets

    Args:
        manager (DeviceManager): A device manager provided
            by iotile-gateway.
        loop (IOLoop): A tornado IOLoop that this agent
            shold integrate into.
        args (dict): A dictionary of arguments for configuring
            this agent.
    """

    def __init__(self, args, manager, loop):
        self._args = args
        self.app = None
        self._manager = manager
        self._loop = loop
        self._logger = logging.getLogger(__name__)
        self._logger.addHandler(logging.NullHandler())
        self._logger.setLevel(logging.INFO)

    def start(self):
        """Start this gateway agent

        Called before the event loop is running
        """

        port = self._args.get('port', 5120)

        self.app = tornado.web.Application([
            (r'/iotile/v1', WebSocketHandler, {'manager': self._manager}),
        ])

        self._logger.info("Starting Websocket Agent on port %d" % port)
        self.app.listen(port)

    def stop(self):
        """Stop this gateway agent

        Called with the event loop running
        """

        pass



"""A websocket client that replicates the state of the supervisor."""

from iotile.core.utilities.validating_wsclient import ValidatingWSClient
from iotile.core.exceptions import ArgumentError
from monotonic import monotonic
from threading import Lock
from copy import copy
import command_formats
import states


class ServiceStatusClient(ValidatingWSClient):
    """A websocket client that syncs the state of all known services.

    On creation it connects to the supervisor service and gets the
    current status of each known service.  Then it listens for
    change notifications and updates it internal map of service states.

    Each service is assigned a unique number based on the order in
    which it was registered.
    """

    def __init__(self, url, logger_name=__file__):
        """Constructor.

        Args:
            url (string): The URL of the websocket server that we want
                to connect to.
            logger_name (string): An optional name that errors are logged to
        """
        super(ServiceStatusClient, self).__init__(url, logger_name)

        self._state_lock = Lock()
        self.services = {}
        self._name_map = {}
        self._on_change_callback = None

        # Register callbacks for all of the status notifications
        self.add_message_type(command_formats.ServiceStatusChanged, self._on_status_change)
        self.add_message_type(command_formats.ServiceAdded, self._on_service_added)
        self.add_message_type(command_formats.HeartbeatReceived, self._on_heartbeat)
        self.add_message_type(command_formats.NewMessage, self._on_message)
        self.add_message_type(command_formats.NewHeadline, self._on_headline)
        self.start()

        with self._state_lock:
            self.services = self.sync_services()
            for i, name in enumerate(self.services.iterkeys()):
                self._name_map[i] = name

    def notify_changes(self, callback):
        """Register to receive a callback when a service changes state.

        Args:
            callback (callable): A function to be called with signature
                callback(short_name, id, state, is_new, new_headline)
        """

        self._on_change_callback = callback

    def local_service(self, name_or_id):
        """Get the locally synced information for a service.

        Args:
            name_or_id (string or int): Either a short name for the service or
                a numeric id.

        Returns:
            ServiceState: the current state of the service synced locally
                at the time of the call.
        """

        with self._state_lock:
            if isinstance(name_or_id, int) or isinstance(name_or_id, long):
                if name_or_id not in self._name_map:
                    raise ArgumentError("Unknown ID used to look up service", id=name_or_id)

                name = self._name_map[name_or_id]
            else:
                name = name_or_id

            if name not in self.services:
                raise ArgumentError("Unknown service name", name=name)

            service = self.services[name]
            return copy(service)

    def local_services(self):
        """Get a list of id, name pairs for all of the known synced services.

        Returns:
            list (id, name): A list of tuples with id and service name sorted by id
                from low to high
        """

        with self._state_lock:
            return sorted([(x, y) for x, y in self._name_map.iteritems()], key=lambda x: x[0])

    def sync_services(self):
        """Poll the current state of all services.

        Returns:
            dict: A dictionary mapping service name to service status
        """

        services = {}

        servs = self.list_services()
        for i, serv in enumerate(servs):
            info = self.service_info(serv)
            status = self.service_status(serv)
            messages = self.get_messages(serv)
            headline = self.get_headline(serv)

            services[serv] = states.ServiceState(info['short_name'], info['long_name'], info['preregistered'], i)
            services[serv].state = status['numeric_status']

            for message in messages:
                services[serv].post_message(message.level, message.message, message.count, message.created)

            if headline is not None:
                services[serv].set_headline(headline.level, headline.message, headline.created)

        return services

    def get_messages(self, name):
        """Get stored messages for a service.

        Args:
            name (string): The name of the service to get messages from.

        Returns:
            list(ServiceMessage): A list of the messages stored for this service
        """

        resp = self.send_command('query_messages', {'name': name}, timeout=5.0)
        return [states.ServiceMessage.FromDictionary(x) for x in resp['payload']]

    def get_headline(self, name):
        """Get stored messages for a service.

        Args:
            name (string): The name of the service to get messages from.

        Returns:
            ServiceMessage: the headline or None if no headline has been set
        """

        resp = self.send_command('query_headline', {'name': name}, timeout=5.0)
        if 'payload' not in resp:
            return None

        msg = resp['payload']
        return states.ServiceMessage.FromDictionary(msg)

    def list_services(self):
        """Get the current list of services from the server.

        Returns:
            list(string): A list of the short service names known
                to the supervisor
        """

        resp = self.send_command('list_services', {}, timeout=5.0)
        return resp['payload']['services']

    def send_heartbeat(self, name):
        """Send a heartbeat for a service.

        Args:
            name (string): The name of the service to send a heartbeat for
        """

        resp = self.send_command('heartbeat', {'name': name}, timeout=5.0)
        if resp['success'] is not True:
            raise ArgumentError("Unknown service name", name=name)

    def service_info(self, name):
        """Pull descriptive info of a service by name.

        Information returned includes the service's user friendly
        name and whether it was preregistered or added dynamically.

        Returns:
            dict: A dictionary of service information with the following keys
                set:
                long_name (string): The user friendly name of the service
                preregistered (bool): Whether the service was explicitly
                    called out as a preregistered service.
        """

        resp = self.send_command('query_info', {'name': name}, timeout=5.0)

        if resp['success'] is not True:
            raise ArgumentError("Unknown service name", name=name)

        return resp['payload']

    def update_state(self, name, state):
        """Update the state for a service.

        Args:
            name (string): The name of the service
            state (int): The new state of the service
        """

        resp = self.send_command('update_state', {'name': name, 'new_status': state}, timeout=5.0)
        if resp['success'] is not True:
            raise ArgumentError("Error updating service state", reason=resp['reason'])

    def post_headline(self, name, level, message):
        """Asynchronously update the sticky headline for a service.

        Args:
            name (string): The name of the service
            level (int): A message level in states.*_LEVEL
            message (string): The user facing error message that will be stored
                for the service and can be queried later.
        """

        now = monotonic()
        self.post_command('set_headline', {'name': name, 'level': level, 'message': message, 'created_time': now, 'now_time': now})

    def post_state(self, name, state):
        """Asynchronously try to update the state for a service.

        If the update fails, nothing is reported because we don't wait for a
        response from the server.  This function will return immmediately and not block.

        Args:
            name (string): The name of the service
            state (int): The new state of the service
        """

        self.post_command('update_state', {'name': name, 'new_status': state})

    def post_error(self, name, message):
        """Asynchronously post a user facing error message about a service.

        Args:
            name (string): The name of the service
            message (string): The user facing error message that will be stored
                for the service and can be queried later.
        """

        self.post_command('post_message', {'name': name, 'level': states.ERROR_LEVEL, 'message': message})

    def post_warning(self, name, message):
        """Asynchronously post a user facing warning message about a service.

        Args:
            name (string): The name of the service
            message (string): The user facing warning message that will be stored
                for the service and can be queried later.
        """

        self.post_command('post_message', {'name': name, 'level': states.WARNING_LEVEL, 'message': message})

    def post_info(self, name, message):
        """Asynchronously post a user facing info message about a service.

        Args:
            name (string): The name of the service
            message (string): The user facing info message that will be stored
                for the service and can be queried later.
        """

        self.post_command('post_message', {'name': name, 'level': states.INFO_LEVEL, 'message': message})

    def service_status(self, name):
        """Pull the current status of a service by name.

        Returns:
            dict: A dictionary of service status
        """

        resp = self.send_command('query_status', {'name': name}, timeout=5.0)

        if resp['success'] is not True:
            raise ArgumentError("Unknown service name", name=name)

        return resp['payload']

    def register_service(self, short_name, long_name, allow_duplicate=True):
        """Register a new service with the service manager.

        Args:
            short_name (string): A unique short name for this service that functions
                as an id
            long_name (string): A user facing name for this service
            allow_duplicate (boolean): Don't throw an error if this service is already
                registered.  This is important if the service is preregistered for example.
        Raises:
            ArgumentError: if the short_name is already taken
        """

        resp = self.send_command('register_service', {'name': short_name, 'long_name': long_name})

        if resp['success'] is not True and not allow_duplicate:
            raise ArgumentError("Service name already registered", short_name=short_name)

    def _on_status_change(self, update):
        """Update a service that has its status updated."""

        info = update['payload']
        new_number = info['new_status']
        name = update['name']
        is_changed = False

        with self._state_lock:
            if name not in self.services:
                return

            is_changed = self.services[name].state != new_number

            self.services[name].state = new_number

        # Notify about this service state change if anyone is listening
        if self._on_change_callback and is_changed:
            self._on_change_callback(name, self.services[name].id, new_number, False, False)

    def _on_service_added(self, update):
        """Add a new service."""

        info = update['payload']
        name = info['short_name']

        with self._state_lock:
            if name in self.services:
                return

            new_id = len(self.services)
            serv = states.ServiceState(name, info['long_name'], info['preregistered'], new_id)
            self.services[name] = serv
            self._name_map[new_id] = name

        # Notify about this new service if anyone is listening
        if self._on_change_callback:
            self._on_change_callback(name, new_id, serv.state, True, False)

    def _on_heartbeat(self, update):
        """Receive a new heartbeat for a service."""

        name = update['name']

        with self._state_lock:
            if name not in self.services:
                return

            self.services[name].heartbeat()

    def _on_message(self, update):
        """Receive a message from a service."""

        name = update['name']
        message_obj = update['payload']

        with self._state_lock:
            if name not in self.services:
                return

            self.services[name].post_message(message_obj['level'], message_obj['message'])

    def _on_headline(self, update):
        """Receive a headline from a service."""

        name = update['name']
        message_obj = update['payload']
        new_headline = False

        with self._state_lock:
            if name not in self.services:
                return

            self.services[name].set_headline(message_obj['level'], message_obj['message'])

            if self.services[name].headline.count == 1:
                new_headline = True
        # Notify about this service state change if anyone is listening
        # headline changes are only reported if they are not duplicates
        if self._on_change_callback and new_headline:
            self._on_change_callback(name, self.services[name].id, self.services[name].state, False, True)



"""A websocket handler for managing running services."""

import logging
import datetime
import tornado.websocket
import tornado.gen
import msgpack
from command_formats import CommandMessage
from iotile.core.exceptions import ArgumentError, ValidationError


class ServiceWebSocketHandler(tornado.websocket.WebSocketHandler):
    """A websocket interface to ServiceManager."""

    def initialize(self, manager, logger):
        """Initialize this handler."""
        self.manager = manager
        self.logger = logger

    @classmethod
    def decode_datetime(cls, obj):
        """Decode a datetime from msgpack."""
        if b'__datetime__' in obj:
            obj = datetime.datetime.strptime(obj[b'as_str'].decode(), "%Y%m%dT%H:%M:%S.%f")
        return obj

    @classmethod
    def encode_datetime(cls, obj):
        """Encode a datetime for msgpack."""
        if isinstance(obj, datetime.datetime):
            obj = {'__datetime__': True, 'as_str': obj.strftime("%Y%m%dT%H:%M:%S.%f").encode()}
        return obj

    def unpack(self, message):
        """Unpack a binary message packed message with datetime handling."""
        return msgpack.unpackb(message, object_hook=self.decode_datetime)

    def on_message(self, message):
        """Handle a message received on the websocket."""
        cmd = self.unpack(message)

        try:
            CommandMessage.verify(cmd)
            self._on_command(cmd)
        except ValidationError:
            if 'operation' in cmd:
                print("Invalid operation received: %s" % cmd['operation'])
            self.send_error('message did not correspond with a known schema')

    def _on_command(self, cmd):
        """Process a command to the status server.

        The command is previously validated in on_message
        """

        op = cmd['operation']

        del cmd['operation']
        del cmd['type']
        self.logger.info("Received %s with payload %s", op, cmd)

        if op == 'heartbeat':
            try:
                self.manager.send_heartbeat(cmd['name'])

                if not cmd['no_response']:
                    self.send_response(True, None)
            except Exception, exc:
                if not cmd['no_response']:
                    self.send_error(str(exc))
        elif op == 'list_services':
            names = self.manager.list_services()
            if not cmd['no_response']:
                self.send_response(True, {'services': names})
        elif op == 'query_status':
            try:
                status = self.manager.service_status(cmd['name'])
                if not cmd['no_response']:
                    self.send_response(True, status)
            except ArgumentError:
                if not cmd['no_response']:
                    self.send_error("Service name could not be found")
        elif op == 'register_service':
            try:
                status = self.manager.add_service(cmd['name'], cmd['long_name'])
                if not cmd['no_response']:
                    self.send_response(True, None)
            except ArgumentError:
                if not cmd['no_response']:
                    self.send_error("Service was already registered")
        elif op == 'query_info':
            try:
                info = self.manager.service_info(cmd['name'])
                if not cmd['no_response']:
                    self.send_response(True, info)
            except ArgumentError:
                if not cmd['no_response']:
                    self.send_error("Service name could not be found")
        elif op == 'query_messages':
            try:
                msgs = self.manager.service_messages(cmd['name'])
                if not cmd['no_response']:
                    self.send_response(True, [msg.to_dict() for msg in msgs])
            except ArgumentError:
                if not cmd['no_response']:
                    self.send_error("Service name could not be found")
        elif op == 'query_headline':
            try:
                headline = self.manager.service_headline(cmd['name'])
                if not cmd['no_response']:
                    if headline is not None:
                        headline = headline.to_dict()
                    self.send_response(True, headline)
            except ArgumentError:
                if not cmd['no_response']:
                    self.send_error("Service name could not be found")
        elif op == 'update_state':
            try:
                self.manager.update_state(cmd['name'], cmd['new_status'])
                if not cmd['no_response']:
                    self.send_response(True, None)
            except ArgumentError, exc:
                if not cmd['no_response']:
                    self.send_error(str(exc))
        elif op == 'post_message':
            try:
                self.manager.send_message(cmd['name'], cmd['level'], cmd['message'])
                if not cmd['no_response']:
                    self.send_response(True, None)
            except ArgumentError, exc:
                if not cmd['no_response']:
                    self.send_error(str(exc))
        elif op == 'set_headline':
            try:
                self.manager.set_headline(cmd['name'], cmd['level'], cmd['message'])
                if not cmd['no_response']:
                    self.send_response(True, None)
            except ArgumentError, exc:
                if not cmd['no_response']:
                    self.send_error(str(exc))
        else:
            if not cmd['no_response']:
                self.send_error("Unknown command: %s" % op)

    def send_response(self, success, obj):
        """Send a response back to someone."""

        resp_object = {'type': 'response', 'success': success}
        if obj is not None:
            resp_object['payload'] = obj

        msg = msgpack.packb(resp_object, default=self.encode_datetime)

        try:
            self.write_message(msg, binary=True)
        except tornado.websocket.WebSocketClosedError:
            pass

    def send_error(self, reason):
        """Send an error to someone."""

        msg = msgpack.packb({'type': 'response', 'success': False, 'reason': reason})

        try:
            self.write_message(msg, binary=True)
        except tornado.websocket.WebSocketClosedError:
            pass

    def send_notification(self, name, change_type, change_info):
        """Send an unsolicited notification to someone."""

        notif_object = {'type': 'notification', 'operation': change_type, 'name': name}
        if change_info is not None:
            notif_object['payload'] = change_info

        msg = msgpack.packb(notif_object)

        try:
            self.write_message(msg, binary=True)
        except tornado.websocket.WebSocketClosedError:
            pass

    def open(self, *args):
        """Register that someone opened a connection."""
        self.stream.set_nodelay(True)
        self.manager.add_monitor(self.send_notification)
        self.logger.info('Client connected')

    def on_close(self, *args):
        """Register that someone closed a connection."""

        self.manager.remove_monitor(self.send_notification)
        self.logger.info('Client disconnected')
